/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the papyrus library.                             *
 *                                                                         *
 *   papyrus is free software; you can redistribute it and/or modify       *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 3.0 as published by the Free Software Foundation.             *
 *                                                                         *
 *   papyrus is distributed in the hope that it will be useful,            *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License version 3.0 for more details.       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with the papyrus library. If not, see                   *
 *   <http://www.gnu.org/licenses/>.                                       *
 ***************************************************************************/
#ifndef PAPYRUSRENDERABLE_H
#define PAPYRUSRENDERABLE_H

#include <map>
#include <glibmm/ustring.h>
#include <papyrus/pointer.h>

#include <papyrus/object.h>
#include <papyrus/utility.h>
#include <cairomm/cairomm.h>

namespace Papyrus
{

  /**
   * Papyrus::Renderable is a common base for all classes capable of rendering themselves
   * into a cairomm context.
   *
   * Renderable provides the pure virtual \c render() method, which takes a
   * cairomm context as its only parameter. Descendants of Papyrus::Renderable
   * provide their own implementations of \c render() that allow them to be
   * rendered into the supplied cairomm context.
   *
   * The \c render() method makes it possible to render a hierarchy (or hierarchies)
   * of Renderables into a cairomm context employing a visitor pattern.
   *
   * <h2>Inheriting from Papyrus::Renderable</h2>
   *
   * <h3>When:</h3>
   * Inherit from Papyrus::Renderable when you want your object to be capable of being
   * rendered into a cairo context through a common interface.
   *
   * <h3>Tips:</h3>
   * Don't forget to provide your own implementation of \c render() since it's
   * a pure virtual method.
   *
   * @author Rick L Vinyard Jr
   */
  class Renderable: public Object
  {
    protected:
      Renderable ( const Glib::ustring& id ) : Object ( id )
      { }

    public:
      typedef PapyrusPointer<Renderable> pointer;

      virtual ~Renderable()
      { }

      virtual void render ( Cairo::RefPtr<Cairo::Context> cairo ) const = 0;

      virtual void render( Cairo::RefPtr<Cairo::Context> cairo, double x, double y, double w, double h ) const { this->render(cairo); }

      /**
       * Returns an SVG representation of the Renderable
       */
      virtual Glib::ustring svg(unsigned depth=0) /*= 0;*/ { return Glib::ustring(); }

      PAPYRUS_CLASS_NAME ( "Renderable" );

    protected:
  
      Glib::ustring svg_spacing(unsigned depth)
      {
        Glib::ustring s;
        for ( unsigned i = 0; i < depth; i++ )
          s += "  ";
        return s;
      }

      Glib::ustring svg_id()
      {
        Glib::ustring s;
        if ( ! m_id.empty() )
          s += "id=\"" + m_id + "\" ";
        return s;
      }


  };

};

#endif
